
# Standard library
from copy import deepcopy

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.utils.helpers import set_param, get_param


@pytest.mark.usefixtures('get_individuals', 'root_ffield', 'param_keys')
def test_set_param(get_individuals, root_ffield, param_keys):
    param = get_individuals[2].params[0]
    original_ffield = deepcopy(root_ffield)
    key = param_keys[0]
    set_param(key, root_ffield, param)
    assert root_ffield[2][0][0] == 1.24
    assert root_ffield[2][0][0] != original_ffield[2][0][0]
    # Assure no other parameters are affected by the single change
    for (k1, v1), (k2, v2) in zip(root_ffield.items(), original_ffield.items()):
        if k1 != 2:
            assert v1 == v2
        else:
            for val1, val2 in zip(v1[1:], v2[1:]):
                assert val1 == val2


@pytest.mark.usefixtures('root_ffield', 'param_keys')
def test_get_param(root_ffield, param_keys):
    key = param_keys[0]
    param = get_param(key, root_ffield)
    assert param == root_ffield[2][0][0]
    assert param == 1



#!/usr/bin/env python

"""
Module that contains interface for adaptive genetic algorithm methods to be used for an Individual.

__author__ = "Chad Daksha"
"""

# Standard library
import abc
from typing import List, Tuple

# 3rd party packages

# Local source


class IAdaptationStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def adaptation(average_cost: float, minimum_cost: float, parent_costs: Tuple[float, float], **kwargs) \
            -> Tuple[float, List[float]]:
        raise NotImplementedError



#!/usr/bin/env python

"""Module with interface for computation of objective function."""

# Standard library
import abc

# 3rd party packages

# Local source


class IErrorStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def error(reax_val, dft_val, weight, **kwargs) -> float:
        raise NotImplementedError




# Standard library

# 3rd party packages

# Local source
from domain.adaptation.factory import adaptation_factory
from domain.adaptation.xiao import XiaoAdapt
from domain.adaptation.srinivas import SrinivasAdapt


def test_get_xiao_adaptation():
    xiao_adapt = adaptation_factory('xiao')
    assert xiao_adapt == XiaoAdapt


def test_get_srinivas_adaptation():
    srinivas_adapt = adaptation_factory('srinivas')
    assert srinivas_adapt == SrinivasAdapt





# Standard library

# 3rd party packages

# Local source
from domain.cost.factory import error_calculator_factory
from domain.cost.reax_error import ReaxError


def test_get_reax_error():
    reax_error_calculator = error_calculator_factory('reax_error')
    assert reax_error_calculator == ReaxError




#!/usr/bin/env python

"""
Module that contains interface for mutation methods to be used for an Individual.

__author__ = "Chad Daksha"
"""

# Standard library
import abc

# 3rd party packages


# Local source
from domain.individual import Individual
from domain.root_individual import RootIndividual


class IMutationStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def mutation(parent: Individual, root_individual: RootIndividual, **kwargs) -> Individual:
        raise NotImplementedError



#!/usr/bin/env python

"""Module with densely-connected Feed-Forward Neural Network with one hidden layer."""

# Standard library
from typing import List

# 3rd party packages
import tensorflow as tf

# Local source
from domain.individual import Individual
from domain.root_individual import RootIndividual
from domain.cost.strategy import IErrorStrategy
from domain.neural_network.extract_data import (individuals_to_features_df,
                                                individuals_to_features_and_outputs_df,
                                                extract_features_and_outputs_from_combined_df)
from domain.neural_network.transform_data import (train_test_split, normalize_features,
                                                  get_columns_to_remove,
                                                  remove_problematic_columns)


class FeedForwardNet:

    def __init__(self, population: List[Individual], verbosity: int = 2, train_fraction: float = 0.80,
                 num_epochs: int = 20000):
        self.population = population
        self.num_input_nodes = len(population[0].params)
        self.num_output_nodes = len(population[0].reax_energies)

        self.train_fraction = train_fraction
        self.verbosity = verbosity
        self.num_epochs = num_epochs

        x_y_df = individuals_to_features_and_outputs_df(population)
        train_df, test_df = train_test_split(x_y_df, train_fraction)
        train_x, self.train_y = extract_features_and_outputs_from_combined_df(train_df, self.num_input_nodes)
        test_x, self.test_y = extract_features_and_outputs_from_combined_df(test_df, self.num_input_nodes)
        self.columns_to_remove = get_columns_to_remove(train_x)
        self.train_x = remove_problematic_columns(train_x, self.columns_to_remove)
        self.test_x = remove_problematic_columns(test_x, self.columns_to_remove)

        self.train_stats = self.train_x.describe().transpose()
        self.normalized_train_x = normalize_features(self.train_x, self.train_stats)
        self.normalized_test_x = normalize_features(self.test_x, self.train_stats)

    def execute(self):
        model = self.build()
        history = self.fit(model)
        return model, history

    def build(self):
        model = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(int(self.num_input_nodes), input_shape=[self.num_input_nodes], activation='relu'),
                tf.keras.layers.Dropout(rate=0.2),  # Dropout on the visible (input) layer - rate = frac. inputs to drop
                tf.keras.layers.Dense(self.num_output_nodes, activation='linear',
                                      kernel_regularizer=tf.keras.regularizers.l2(0.1)
                                      )
            ]
        )
        optimizer = tf.keras.optimizers.RMSprop(0.001)
        model.compile(loss='mse',
                      optimizer=optimizer,
                      metrics=[r_square, 'mse'])
        return model

    def fit(self, model):
        """Perform fitting for ANN model to [x, y] data and return its history."""
        # The patience parameter is the amount of epochs to check for improvement
        early_stop = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=50)
        history = model.fit(self.normalized_train_x, self.train_y, epochs=self.num_epochs, batch_size=32,
                            validation_split=0.2, verbose=self.verbosity,
                            callbacks=[
                                early_stop,
                            ]
                            )
        return history

    def predict_outputs(self, model, population: List[Individual]):
        """Given an ANN model and a population, predict the associated outputs with that population."""
        x_df = individuals_to_features_df(population)
        cleaned_x_df = remove_problematic_columns(x_df, self.columns_to_remove)
        normalized_x_df = normalize_features(cleaned_x_df, self.train_stats)
        y_predicted = model.predict(normalized_x_df)
        return y_predicted

    @staticmethod
    def compute_costs(y_predicted, root_individual: RootIndividual, error_strategy: IErrorStrategy, **kwargs) \
            -> List[float]:
        dft_energies = root_individual.dft_energies
        weights = root_individual.weights
        costs = [sum(error_strategy.error(reax_pred, dft_energy, weight, **kwargs)
                     for reax_pred, dft_energy, weight in zip(y_pred_row, dft_energies, weights))
                 for y_pred_row in y_predicted]
        return costs


def r_square(y_true, y_pred):
    """Coefficient of determination (R^2) for regression - only for Keras tensors.
    To be used in metrics args.
    """
    ss_residual = tf.keras.backend.sum(tf.keras.backend.square(y_true - y_pred))
    ss_total = tf.keras.backend.sum(tf.keras.backend.square(y_true - tf.keras.backend.mean(y_true)))
    return 1 - ss_residual / (ss_total + tf.keras.backend.epsilon())



#!/usr/bin/env python

"""Module for custom exceptions that can be raised upon errors."""

# Standard library

# 3rd party packages

# Local source


class ApplicationError(Exception):
    """Base class for other exceptions."""
    pass





import pytest
from tests.fixtures.domain import get_individuals
from tests.fixtures.domain import param_bounds
from tests.fixtures.domain import reax_energies
from tests.fixtures.domain import dft_energies
from tests.fixtures.domain import weights
from tests.fixtures.domain import root_ffield
from tests.fixtures.domain import param_keys

from tests.fixtures.infrastructure import training_set_dir_path
from tests.fixtures.infrastructure import reax_output_dir_path
from tests.fixtures.infrastructure import presenter_output_path
from tests.fixtures.infrastructure import cli_output_path



#!/usr/bin/env python

"""Factory for selection algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from domain.selection.tournament import TournamentSelect


def selection_factory(algorithm_name: str):
    """Factory to select selection type."""
    selection_types = {
        'tournament': TournamentSelect
    }
    return selection_types[algorithm_name]



#!/usr/bin/env python

"""Combines domain logic to propagate (generational) genetic algorithm."""

# Standard library
from typing import List, Tuple
import random
import statistics

# 3rd party packages

# Local source
from domain.individual import Individual
from use_case.port.settings_repository import IAllSettings
from use_case.port.population_repository import IPopulationRepository


class PopulationPropagator:

    def __init__(self, settings_repository: IAllSettings, population_repository: IPopulationRepository):
        self.ga_settings = settings_repository.ga_settings

        self.crossover_rate = self.ga_settings.crossover_rate
        self.mutation_rates = [self.ga_settings.mutation_rate, self.ga_settings.mutation_rate]

        self.mutation_strategy = settings_repository.strategy_settings.mutation_strategy
        self.crossover_strategy = settings_repository.strategy_settings.crossover_strategy
        self.selection_strategy = settings_repository.strategy_settings.selection_strategy
        self.adaptation_strategy = settings_repository.strategy_settings.adaptation_strategy

        self.ga_settings_dict = vars(settings_repository.ga_settings)
        self.mutation_settings_dict = vars(settings_repository.mutation_settings)
        self.crossover_settings_dict = vars(settings_repository.crossover_settings)
        self.selection_settings_dict = vars(settings_repository.selection_settings)
        self.adaptation_settings_dict = vars(settings_repository.adaptation_settings)

        self.root_individual = population_repository.get_root_individual()

    def execute(self, parents: List[Individual]) -> List[Individual]:
        """Create next generation using parents."""
        average_cost, minimum_cost = self.compute_statistics(parents)

        children = self.initialize(parents)
        while len(children) < self.ga_settings.population_size:
            parent1 = self.select(parents)
            parent2 = self.select(parents)

            self.adapt_cross_and_mutate_rates(average_cost, minimum_cost, (parent1.cost, parent2.cost))

            child1, child2 = self.cross(parent1, parent2)
            child1, child2 = self.mutate(child1, child2)

            children.extend([child1, child2])

        return children

    def initialize(self, parents: List[Individual]) -> List[Individual]:
        children = []
        if self.ga_settings.use_elitism:
            children.extend(sorted(parents)[0:2])
        return children

    def select(self, parents: List[Individual]) -> Individual:
        return self.selection_strategy.selection(parents, **self.selection_settings_dict)

    def cross(self, parent1: Individual, parent2: Individual) -> Tuple[Individual, Individual]:
        child1, child2 = parent1, parent2
        if random.random() < self.crossover_rate:
            child1, child2 = self.crossover_strategy.crossover(parent1, parent2, self.root_individual,
                                                               **self.crossover_settings_dict)
        return child1, child2

    def mutate(self, parent1: Individual, parent2: Individual) -> Tuple[Individual, Individual]:
        child1, child2 = parent1, parent2
        if random.random() < self.mutation_rates[0]:
            child1 = self.mutation_strategy.mutation(parent1, self.root_individual, **self.mutation_settings_dict)
        if random.random() < self.mutation_rates[1]:
            child2 = self.mutation_strategy.mutation(parent2, self.root_individual, **self.mutation_settings_dict)
        return child1, child2

    def adapt_cross_and_mutate_rates(self, average_cost: float, minimum_cost: float, parent_costs: Tuple[float, float]):
        if self.ga_settings.use_adaptation:
            self.crossover_rate, self.mutation_rates = \
                self.adaptation_strategy.adaptation(average_cost, minimum_cost, parent_costs,
                                                    **self.adaptation_settings_dict, **self.ga_settings_dict)

    @staticmethod
    def compute_statistics(population: List[Individual]) -> Tuple[float, float]:
        costs = [individual.cost for individual in population]
        return statistics.mean(costs), min(costs)



#!/usr/bin/env python

"""Factory for error calculation algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from domain.cost.reax_error import ReaxError


def error_calculator_factory(algorithm_name: str):
    """Factory to select cost/error calculator type."""
    error_calculator_types = {
        'reax_error': ReaxError,
    }
    return error_calculator_types[algorithm_name]



#!/usr/bin/env python

# Standard library
from typing import Tuple
import random

# 3rd party packages


# Local source
from domain.crossover.strategy import ICrossoverStrategy
from domain.individual import Individual
from domain.root_individual import RootIndividual


class TwoPointCross(ICrossoverStrategy):

    @staticmethod
    def crossover(parent1: Individual, parent2: Individual, root_individual: RootIndividual,
                  **kwargs) -> Tuple[Individual, Individual]:
        """Execute two-point crossover."""
        idx = random.sample(len(parent1.params), 2)
        smaller_id = min(idx)
        bigger_id = max(idx)
        child1 = Individual(parent1.params[:smaller_id] + parent2.params[smaller_id:bigger_id]
                            + parent1.params[bigger_id:], root_individual=root_individual)
        child2 = Individual(parent2.params[:smaller_id] + parent1.params[smaller_id:bigger_id]
                            + parent2.params[bigger_id:], root_individual=root_individual)
        return child1, child2




# Standard library
from unittest.mock import patch

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.mutation.gauss import GaussianMutate
from tests.domain.crossover.test_double_pareto import get_root_individual


@pytest.mark.usefixtures('get_individuals')
def test_single_gauss_with_frac(get_root_individual, get_individuals):
    parent = get_individuals[0]
    child = GaussianMutate.mutation(parent, get_root_individual)
    assert parent.params != child.params
    assert parent.params == [1.0, 0.1, -0.5, 4.3, 2.4]


@pytest.mark.usefixtures('get_individuals')
def test_multi_gauss(get_root_individual, get_individuals):
    parent = get_individuals[0]
    child = GaussianMutate.mutation(parent, root_individual=get_root_individual,
                                    gauss_std=[0.01, 0.1, 1.0], gauss_frac=[0.25, 0.5, 0.25])
    assert parent.params != child.params
    assert parent.params == get_individuals[0].params


@patch('random.gauss')
@patch('random.uniform')
@pytest.mark.usefixtures('get_individuals')
def test_multi_gauss_std_1(uniform_mock, gauss_mock, get_root_individual, get_individuals):
    uniform_mock.return_value = 0.15
    parent = get_individuals[0]
    GaussianMutate.mutation(parent, root_individual=get_root_individual,
                            gauss_std=[0.01, 0.1, 1.0], gauss_frac=[0.25, 0.5, 0.25])
    gauss_mock.assert_called_with(0, 0.01)


@patch('random.gauss')
@patch('random.uniform')
@pytest.mark.usefixtures('get_individuals')
def test_multi_gauss_std_2(uniform_mock, gauss_mock, get_root_individual, get_individuals):
    uniform_mock.return_value = 0.50
    parent = get_individuals[0]
    GaussianMutate.mutation(parent, root_individual=get_root_individual,
                            gauss_std=[0.01, 0.1, 1.0], gauss_frac=[0.25, 0.5, 0.25])
    gauss_mock.assert_called_with(0, 0.1)


@patch('random.gauss')
@patch('random.uniform')
@pytest.mark.usefixtures('get_individuals')
def test_multi_gauss_std_1(uniform_mock, gauss_mock, get_root_individual, get_individuals):
    uniform_mock.return_value = 0.85
    parent = get_individuals[0]
    GaussianMutate.mutation(parent, root_individual=get_root_individual,
                            gauss_std=[0.01, 0.1, 1.0], gauss_frac=[0.25, 0.5, 0.25])
    gauss_mock.assert_called_with(0, 1.0)




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual
from parametrization_clean.domain.cost.reax_error import ReaxError


@pytest.fixture()
@pytest.mark.usefixtures('dft_energies', 'weights', 'root_ffield', 'param_keys')
def root_individual(dft_energies, weights, root_ffield, param_keys):
    root_individual = RootIndividual(dft_energies, weights, root_ffield, param_keys)
    return root_individual


@pytest.fixture()
def individual():
    params = [0.5, 0.4, 1.0, -4.7, 8.6]
    reax_energies = [43.2, 10.6, -174.2, 1008.4]
    individual = Individual(params=params, reax_energies=reax_energies, error_calculator=ReaxError)
    return individual


def test_individual_init():
    params = [0.5, 0.4, 1.0, -4.7, 8.6]
    reax_energies = [43.2, 10.6, -174.2, 1008.4]
    individual = Individual(params=params, reax_energies=reax_energies)
    assert individual.params == params
    assert individual.reax_energies == reax_energies


def test_individual_total_error(individual, root_individual):
    assert abs(individual.total_error(root_individual) - 871.9892) <= 1e-4


def test_individual_update_ffield(individual, root_individual):
    individual.ffield = individual.update_ffield(root_individual)
    assert individual.ffield != root_individual.root_ffield
    assert individual.ffield[2][0][0] == 0.5
    assert individual.ffield[2][3][0] == 0.4
    assert individual.ffield[3][0][0] == 1.0
    assert individual.ffield[3][0][2] == -4.7
    assert individual.ffield[3][2][1] == 8.6
    assert root_individual.root_ffield[2][0][0] == 1
    assert root_individual.root_ffield[2][3][0] == 7
    assert root_individual.root_ffield[3][0][0] == 1
    assert root_individual.root_ffield[3][0][2] == 3
    assert root_individual.root_ffield[3][2][1] == 8
    # Assure no other parameters are affected by the changes
    for (k1, v1), (k2, v2) in zip(individual.ffield.items(), root_individual.root_ffield.items()):
        if k1 == 1:
            assert v1 == v2
        elif k1 == 2:
            for i in range(len(v1)):
                dont_check = {0, 3}
                if i not in dont_check:
                    assert v1[i] == v2[i]
                else:
                    for j in range(len(v1[i])):
                        dont_check = {0}
                        if j not in dont_check:
                            assert v1[i][j] == v2[i][j]
        else:
            for i in range(len(v1)):
                dont_check = {0, 2}
                if i not in dont_check:
                    assert v1[i] == v2[i]
                else:
                    for j in range(len(v1[i])):
                        dont_check = {0, 1, 2}
                        if j not in dont_check:
                            assert v1[i][j] == v2[i][j]


def test_individual_from_root_individual(root_individual):
    individual_from_root = Individual.from_root_individual(root_individual)
    assert isinstance(individual_from_root, Individual)
    assert individual_from_root.params == root_individual.root_params
    assert individual_from_root.reax_energies is None
    assert individual_from_root.ffield is not None
    assert individual_from_root.cost is None


@pytest.mark.usefixtures('get_individuals')
def test_individual_eq(get_individuals):
    get_individuals[0].cost = 5142
    get_individuals[1].cost = 4387
    get_individuals[2].cost = 5142
    assert get_individuals[0] == get_individuals[2]
    assert get_individuals[0] != get_individuals[1]


@pytest.mark.usefixtures('get_individuals')
def test_individual_lt(get_individuals):
    get_individuals[0].cost = 5142
    get_individuals[1].cost = 4387
    assert get_individuals[1] < get_individuals[0]


@pytest.mark.usefixtures('get_individuals')
def test_individual_gt(get_individuals):
    get_individuals[0].cost = 5142
    get_individuals[1].cost = 4387
    assert get_individuals[0] > get_individuals[1]



#!/usr/bin/env python

"""Default parameters used for genetic algorithm/neural network."""

# Standard library
from typing import List
# 3rd party packages

# Local source
from domain.selection.tournament import TournamentSelect
from domain.mutation.gauss import GaussianMutate
from domain.crossover.double_pareto import DoubleParetoCross
from domain.adaptation.xiao import XiaoAdapt
from domain.cost.reax_error import ReaxError
from domain.mutation.nakata import NakataMutate
from use_case.port.settings_repository import (IStrategySettings,
                                               IGeneticAlgorithmSettings,
                                               IMutationSettings,
                                               ICrossoverSettings,
                                               ISelectionSettings,
                                               IAdaptationSettings,
                                               INeuralNetSettings,
                                               IAllSettings)


class DefaultSettings(IAllSettings):

    def __init__(self):
        self._strategy_settings = DefaultStrategySettings()
        self._ga_settings = DefaultGeneticAlgorithmSettings()
        self._mutation_settings = DefaultMutationSettings()
        self._crossover_settings = DefaultCrossoverSettings()
        self._selection_settings = DefaultSelectionSettings()
        self._adaptation_settings = DefaultAdaptationSettings()
        self._neural_net_settings = DefaultNeuralNetSettings()

    @property
    def strategy_settings(self) -> IStrategySettings:
        return self._strategy_settings

    @property
    def ga_settings(self) -> IGeneticAlgorithmSettings:
        return self._ga_settings

    @property
    def mutation_settings(self) -> IMutationSettings:
        return self._mutation_settings

    @property
    def crossover_settings(self) -> ICrossoverSettings:
        return self._crossover_settings

    @property
    def selection_settings(self) -> ISelectionSettings:
        return self._selection_settings

    @property
    def adaptation_settings(self) -> IAdaptationSettings:
        return self._adaptation_settings

    @property
    def neural_net_settings(self) -> INeuralNetSettings:
        return self._neural_net_settings


class DefaultStrategySettings(IStrategySettings):

    def __init__(self):
        self._selection_strategy = TournamentSelect
        self._mutation_strategy = GaussianMutate
        self._crossover_strategy = DoubleParetoCross
        self._adaptation_strategy = XiaoAdapt
        self._error_strategy = ReaxError
        self._initialization_strategy = NakataMutate

    @property
    def selection_strategy(self):
        return self._selection_strategy

    @property
    def mutation_strategy(self):
        return self._mutation_strategy

    @property
    def crossover_strategy(self):
        return self._crossover_strategy

    @property
    def adaptation_strategy(self):
        return self._adaptation_strategy

    @property
    def error_strategy(self):
        return self._error_strategy

    @property
    def initialization_strategy(self):
        return self._initialization_strategy


class DefaultGeneticAlgorithmSettings(IGeneticAlgorithmSettings):

    def __init__(self):
        self._population_size = 30
        self._mutation_rate = 0.2
        self._crossover_rate = 0.8
        self._use_elitism = True
        self._use_adaptation = False
        self._use_neural_network = False

    @property
    def population_size(self) -> int:
        return self._population_size

    @property
    def mutation_rate(self) -> float:
        return self._mutation_rate

    @property
    def crossover_rate(self) -> float:
        return self._crossover_rate

    @property
    def use_elitism(self) -> bool:
        return self._use_elitism

    @property
    def use_adaptation(self) -> bool:
        return self._use_adaptation

    @property
    def use_neural_network(self) -> bool:
        return self._use_neural_network


class DefaultMutationSettings(IMutationSettings):

    def __init__(self):
        self._gauss_std = [0.01, 0.1]
        self._gauss_frac = [0.5, 0.5]
        self._nakata_rand_lower = -1.0
        self._nakata_rand_higher = 1.0
        self._nakata_scale = 0.1
        self._polynomial_eta = 60
        self._param_bounds = []

    @property
    def gauss_std(self) -> List[float]:
        return self._gauss_std

    @property
    def gauss_frac(self) -> List[float]:
        return self._gauss_frac

    @property
    def nakata_rand_lower(self) -> float:
        return self._nakata_rand_lower

    @property
    def nakata_rand_higher(self) -> float:
        return self._nakata_rand_higher

    @property
    def nakata_scale(self) -> float:
        return self._nakata_scale

    @property
    def polynomial_eta(self) -> float:
        return self._polynomial_eta

    @property
    def param_bounds(self):
        return self._param_bounds


class DefaultCrossoverSettings(ICrossoverSettings):

    def __init__(self):
        self._dpx_alpha = 10
        self._dpx_beta = 1

    @property
    def dpx_alpha(self) -> float:
        return self._dpx_alpha

    @property
    def dpx_beta(self) -> float:
        return self._dpx_beta


class DefaultSelectionSettings(ISelectionSettings):

    def __init__(self):
        self._tournament_size = 2

    @property
    def tournament_size(self) -> int:
        return self._tournament_size


class DefaultAdaptationSettings(IAdaptationSettings):

    def __init__(self):
        self._srinivas_k1 = 1.0
        self._srinivas_k2 = 0.5
        self._srinivas_k3 = 1.0
        self._srinivas_k4 = 0.5
        self._srinivas_default_mutation_rate = 0.01
        self._xiao_min_crossover_rate = DefaultGeneticAlgorithmSettings().crossover_rate * 0.8
        self._xiao_min_mutation_rate = DefaultGeneticAlgorithmSettings().mutation_rate * 0.8
        self._xiao_scale = 40

    @property
    def srinivas_k1(self) -> float:
        return self._srinivas_k1

    @property
    def srinivas_k2(self) -> float:
        return self._srinivas_k2

    @property
    def srinivas_k3(self) -> float:
        return self._srinivas_k3

    @property
    def srinivas_k4(self) -> float:
        return self._srinivas_k4

    @property
    def srinivas_default_mutation_rate(self) -> float:
        return self._srinivas_default_mutation_rate

    @property
    def xiao_min_crossover_rate(self) -> float:
        return self._xiao_min_crossover_rate

    @property
    def xiao_min_mutation_rate(self) -> float:
        return self._xiao_min_mutation_rate

    @property
    def xiao_scale(self) -> float:
        return self._xiao_scale


class DefaultNeuralNetSettings(INeuralNetSettings):

    def __init__(self):
        self._verbosity = 2
        self._train_fraction = 0.8
        self._num_epochs = 10000
        self._num_populations_to_train_on = 1
        self._num_nested_ga_iterations = 1
        self._minimum_validation_r_squared = 0.95

    @property
    def verbosity(self) -> int:
        return self._verbosity

    @property
    def train_fraction(self) -> float:
        return self._train_fraction

    @property
    def num_epochs(self) -> int:
        return self._num_epochs

    @property
    def num_populations_to_train_on(self) -> int:
        return self._num_populations_to_train_on

    @property
    def num_nested_ga_iterations(self) -> int:
        return self._num_nested_ga_iterations

    @property
    def minimum_validation_r_squared(self) -> float:
        return self._minimum_validation_r_squared



#!/usr/bin/env python

"""Concrete implementation of population repository interface. Reads files to get data."""

# Standard library
from typing import List, Tuple
from pathlib import Path
import os
import shutil

# 3rd party packages

# Local source
from domain.individual import Individual
from domain.root_individual import RootIndividual
from domain.cost.reax_error import ReaxError
from domain.utils.helpers import get_param
from use_case.port.population_repository import IPopulationRepository
from use_case.port.settings_repository import IAllSettings
from infrastructure.utils.reax_reader import ReaxReader, write_ffield
from infrastructure.utils.reax_converter import Fort99Extractor
from infrastructure.utils.response_object import (ResponseSuccess, ResponseWarning, ResponseFailure)


class PopulationFileRepository(IPopulationRepository):
    GENERATION_FOLDER_PREFIX = "generation-"
    INDIVIDUAL_FOLDER_PREFIX = "child-"

    def __init__(self, training_set_path, population_path, settings_repository: IAllSettings,
                 current_generation_number):
        self.training_set_path = Path(training_set_path)
        self.population_path = Path(population_path)

        self.training_reax_reader = ReaxReader(training_set_path)
        self.population_reax_reader = ReaxReader(population_path)

        self.population_size = settings_repository.ga_settings.population_size
        self.current_generation_number = current_generation_number

        self.param_keys, _, _ = self.training_reax_reader.read_params()

        if not os.path.isdir(population_path):
            os.mkdir(population_path)

        _, self.atom_types = self.training_reax_reader.read_ffield()

    def get_root_individual(self) -> RootIndividual:
        root_ffield, _ = self.training_reax_reader.read_ffield()
        fort99_data = self.training_reax_reader.read_fort99()

        # TODO: PROBLEM - First generation also requires 'fort.99' to be present in training file directory!
        fort99_extractor = Fort99Extractor(fort99_data)
        dft_energies = fort99_extractor.get_dft_energies()
        weights = fort99_extractor.get_weights()
        return RootIndividual(dft_energies, weights, root_ffield, self.param_keys)

    def get_population(self, generation_number: int) -> Tuple[List[Individual], List[int]]:
        generation_dir_path = os.path.join(self.population_path, self.GENERATION_FOLDER_PREFIX + str(generation_number))
        root_individual = self.get_root_individual()

        population = []
        successfully_retrieved_case_numbers = []
        for case_number in range(self.population_size):
            child_dir = os.path.join(generation_dir_path, self.INDIVIDUAL_FOLDER_PREFIX + str(case_number))
            self.population_reax_reader.dir_path = child_dir

            try:
                child_fort99_data = self.population_reax_reader.read_fort99()
                child_ffield, _ = self.population_reax_reader.read_ffield()
                child_params = [get_param(key, child_ffield) for key in self.param_keys]

                fort99_extractor = Fort99Extractor(child_fort99_data)
                child_reax_energies = fort99_extractor.get_reax_energies()

                # noinspection PyTypeChecker
                child = Individual(child_params, child_reax_energies, root_individual, error_calculator=ReaxError)

                population.append(child)
                successfully_retrieved_case_numbers.append(case_number)

            except FileNotFoundError:
                # 'fort.99' file does not exist
                print("fort.99 not found in '{}'...continuing to next case".format(child_dir))
                continue

        return population, successfully_retrieved_case_numbers

    def get_previous_n_populations(self, num_populations: int) -> List[Individual]:
        """Read previous N populations before the current generation number.
        Precaution - ensure that lowest possible generation number is zero.
        """
        lower_bound = max(1, self.current_generation_number - num_populations)
        return self.read_population_range(lower_bound=lower_bound, upper_bound=self.current_generation_number)

    def write_individual(self, individual: Individual, **kwargs):
        """Write a `case` in the population to `child_dir`.

        NOTE: If `child_dir` does not exist, it is created.
        NOTE: control, exe, geo, params, trainset.in files are copied from training set directory to `child_dir`.
        NOTE: '0' is written to `iopt` in `child_dir` due to ReaxFF requirement.
        """
        child_dir = kwargs.get('child_dir', None)
        if not child_dir:
            return ResponseFailure.build_parameters_error(message="Child directory not provided.")

        files_overwritten = True
        if not os.path.isdir(child_dir):
            os.mkdir(child_dir)
            files_overwritten = False

        shutil.copy(os.path.join(self.training_set_path, 'control'), child_dir)
        shutil.copy(os.path.join(self.training_set_path, 'geo'), child_dir)
        shutil.copy(os.path.join(self.training_set_path, 'params'), child_dir)
        shutil.copy(os.path.join(self.training_set_path, 'trainset.in'), child_dir)

        with open(os.path.join(child_dir, 'iopt'), 'w') as out_file:
            out_file.write('0')

        try:
            write_ffield(os.path.join(self.training_set_path, 'ffield'), os.path.join(child_dir, 'ffield'),
                         individual.ffield, self.atom_types)
        except (OSError, IOError):
            return ResponseFailure.build_resource_error(
                message="I/O problem with creating child directory at {}".format(child_dir))
        except ValueError:
            return ResponseFailure.build_resource_error(
                message="Trouble writing ffield file in child directory {}".format(child_dir))

        if files_overwritten:
            return ResponseWarning(message="Files in child directory {} were overwritten."
                                   .format(child_dir))
        else:
            return ResponseSuccess(message="Child successfully written at {}.".format(child_dir))

    def write_population(self, population: List[Individual], generation_number):
        """Write a `population` of individuals to the population generation path."""
        generation_dir_path = os.path.join(self.population_path, self.GENERATION_FOLDER_PREFIX + str(generation_number))

        files_overwritten = True
        if not os.path.isdir(generation_dir_path):
            os.mkdir(generation_dir_path)
            files_overwritten = False

        responses = []
        for i, individual in enumerate(population):
            child_dir = os.path.join(generation_dir_path, self.INDIVIDUAL_FOLDER_PREFIX + str(i))
            responses.append(self.write_individual(individual, child_dir=child_dir))

        if any(responses):
            if files_overwritten:
                return ResponseWarning(message="Population generation directory {} overwritten."
                                       .format(generation_dir_path))
            else:
                return ResponseSuccess(message="Generation successfully written at {}"
                                       .format(generation_dir_path))
        else:
            return ResponseFailure.build_resource_error(
                message="No children in population directory {} were written successfully."
                .format(generation_dir_path))

    def read_population_range(self, lower_bound: int, upper_bound: int) -> List[Individual]:
        """Read population of Individuals from `_generation_path` for generations between
        [lower_bound, upper_bound) (lower bound inclusive, upper bound exclusive).
        """
        population = []
        for generation_number in range(lower_bound, upper_bound):
            generation_population, _ = self.get_population(generation_number)
            population.extend(generation_population)
        return population



#!/usr/bin/env python

"""Module to parse user-provided, local parameters used for genetic algorithm/neural network.
Overrides default parameters.
"""

# Standard library
import os
import sys
import json

# 3rd party packages

# Local source
from infrastructure.config.default import DefaultSettings
from domain.selection.factory import selection_factory
from domain.adaptation.factory import adaptation_factory
from domain.cost.factory import error_calculator_factory
from domain.crossover.factory import crossover_factory
from domain.mutation.factory import mutation_factory


class UserSettings(DefaultSettings):

    PROJECT_ROOT = os.path.abspath(os.path.join(__file__, "../../../../"))
    USER_CONFIG_FILE = os.path.join(PROJECT_ROOT, "config.json")

    def __init__(self):
        super().__init__()

        try:
            with open(self.USER_CONFIG_FILE, 'r') as in_file:
                all_settings_dict = json.load(in_file)
            self.build_from(all_settings_dict)
        except (FileNotFoundError, json.JSONDecodeError):
            # No configuration file provided - using default config values
            # TODO: LOG
            pass

    def build_from(self, all_settings_dict):
        self.set_strategy_settings(all_settings_dict.get('strategy_settings', {}))
        self.set_ga_settings(all_settings_dict.get('ga_settings', {}))
        self.set_mutation_settings(all_settings_dict.get('mutation_settings', {}))
        self.set_crossover_settings(all_settings_dict.get('crossover_settings', {}))
        self.set_selection_settings(all_settings_dict.get('selection_settings', {}))
        self.set_adaptation_settings(all_settings_dict.get('adaptation_settings', {}))
        self.set_neural_net_settings(all_settings_dict.get('neural_net_settings', {}))

    def set_strategy_settings(self, strategy_settings_dict):
        for key, value in strategy_settings_dict.items():
            if key == "initialization":  # initialization and mutation share same factory
                factory_name = "mutation_factory"
            else:
                factory_name = key + "_factory"

            attribute_name = "_" + key + "_strategy"
            factory_to_call = getattr(sys.modules[__name__], factory_name)
            setattr(self._strategy_settings, attribute_name, factory_to_call(value))

    def set_ga_settings(self, ga_settings_dict):
        self.set_attributes_from_json(self._ga_settings, ga_settings_dict)

    def set_mutation_settings(self, mutation_settings_dict):
        self.set_attributes_from_json(self._mutation_settings, mutation_settings_dict)

    def set_crossover_settings(self, crossover_settings_dict):
        self.set_attributes_from_json(self._crossover_settings, crossover_settings_dict)

    def set_selection_settings(self, selection_settings_dict):
        self.set_attributes_from_json(self._selection_settings, selection_settings_dict)

    def set_adaptation_settings(self, adaptation_settings_dict):
        self.set_attributes_from_json(self._adaptation_settings, adaptation_settings_dict)

    def set_neural_net_settings(self, neural_net_settings_dict):
        self.set_attributes_from_json(self._neural_net_settings, neural_net_settings_dict)

    @staticmethod
    def set_attributes_from_json(config_object, json_dict):
        """Extract and set class attributes from JSON dictionary object."""
        for key, value in json_dict.items():
            # adding underscore as each attribute is stored as protected property
            attribute_name = "_" + key
            setattr(config_object, attribute_name, value)



#!/usr/bin/env python

"""Factory for adaptation algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from domain.adaptation.srinivas import SrinivasAdapt
from domain.adaptation.xiao import XiaoAdapt


def adaptation_factory(algorithm_name: str):
    """Factory to select adaptation type."""
    adaptation_types = {
        'srinivas': SrinivasAdapt,
        'xiao': XiaoAdapt,
    }
    return adaptation_types[algorithm_name]




# Standard library

# 3rd party packages

# Local source
from domain.mutation.factory import mutation_factory
from domain.mutation.nakata import NakataMutate
from domain.mutation.central_uniform import CentralUniformMutate
from domain.mutation.polynomial import PolynomialMutate
from domain.mutation.gauss import GaussianMutate


def test_get_nakata():
    assert mutation_factory('nakata') == NakataMutate


def test_get_central_uniform():
    assert mutation_factory('central_uniform') == CentralUniformMutate


def test_get_polynomial():
    assert mutation_factory('polynomial') == PolynomialMutate


def test_get_gauss():
    assert mutation_factory('gauss') == GaussianMutate



#!/usr/bin/env python

"""Response object for repository output usage."""

# Standard library

# 3rd party packages

# Local source


class ResponseSuccess:
    SUCCESS = "SUCCESS"

    def __init__(self, message=None):
        self.type = self.SUCCESS
        self.message = message
        self.status_code = 0

    def __bool__(self):
        return True


class ResponseWarning:
    WARNING = "WARNING"

    def __init__(self, message=None):
        self.type = self.WARNING
        self.message = message
        self.status_code = 0

    def __bool__(self):
        return True


class ResponseFailure:
    RESOURCE_ERROR = "RESOURCE_ERROR"
    PARAMETERS_ERROR = "PARAMETERS_ERROR"
    SYSTEM_ERROR = "SYSTEM_ERROR"

    def __init__(self, type_, message):
        self.type = type_
        self.message = self._format_message(message)
        self.status_code = 1

    @staticmethod
    def _format_message(msg):
        if isinstance(msg, Exception):
            return "{}: {}".format(msg.__class__.__name__, "{}".format(msg))
        return msg

    @property
    def value(self):
        return {'type': self.type, 'message': self.message}

    def __bool__(self):
        return False

    @classmethod
    def build_resource_error(cls, message=None):
        return cls(cls.RESOURCE_ERROR, message)

    @classmethod
    def build_system_error(cls, message=None):
        return cls(cls.SYSTEM_ERROR, message)

    @classmethod
    def build_parameters_error(cls, message=None):
        return cls(cls.PARAMETERS_ERROR, message)



#!/usr/bin/env python

"""Generate output for individuals in the genetic algorithm."""

# Standard library

# 3rd party packages

# Local source
from use_case.port.population_repository import IPopulationRepository


class PopulationWriter:

    def __init__(self, population_repository: IPopulationRepository):
        self.population_repository = population_repository

    def write_individual(self, individual, **kwargs):
        return self.population_repository.write_individual(individual, **kwargs)

    def write_population(self, population, generation_number):
        return self.population_repository.write_population(population, generation_number)




# Standard library

# 3rd party packages

# Local source
from domain.crossover.factory import crossover_factory
from domain.crossover.single_point import SinglePointCross
from domain.crossover.two_point import TwoPointCross
from domain.crossover.uniform import UniformCross
from domain.crossover.double_pareto import DoubleParetoCross


def test_get_single_point():
    single_point_cross = crossover_factory('single_point')
    assert single_point_cross == SinglePointCross


def test_get_two_point():
    assert crossover_factory('two_point') == TwoPointCross


def test_get_uniform():
    assert crossover_factory('uniform') == UniformCross


def test_get_double_pareto():
    assert crossover_factory('double_pareto') == DoubleParetoCross



#!/usr/bin/env python

"""Factory for mutation algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from domain.mutation.nakata import NakataMutate
from domain.mutation.central_uniform import CentralUniformMutate
from domain.mutation.polynomial import PolynomialMutate
from domain.mutation.gauss import GaussianMutate


def mutation_factory(algorithm_name: str):
    """Factory to select mutation type."""
    mutation_types = {
        'nakata': NakataMutate,
        'central_uniform': CentralUniformMutate,
        'polynomial': PolynomialMutate,
        'gauss': GaussianMutate,
    }
    return mutation_types[algorithm_name]




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.cost.reax_error import ReaxError


@pytest.mark.usefixtures('reax_energies', 'dft_energies', 'weights')
def test_reax_error(reax_energies, dft_energies, weights):
    reax_energy = reax_energies[0][0]
    dft_energy = dft_energies[0]
    weight = weights[0]
    true_error = ((reax_energy - dft_energy) / weight) ** 2
    assert ReaxError.error(reax_energy, dft_energy, weight) == true_error
    assert reax_energies[0][0] == reax_energy
    assert dft_energies[0] == dft_energy
    assert weights[0] == weight




# Standard library
from unittest.mock import patch

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.crossover.double_pareto import DoubleParetoCross
from parametrization_clean.domain.root_individual import RootIndividual


@pytest.fixture()
@pytest.mark.usefixtures('dft_energies', 'weights', 'root_ffield', 'param_keys')
def get_root_individual(dft_energies, weights, root_ffield, param_keys):
    return RootIndividual(dft_energies, weights, root_ffield, param_keys)


@pytest.fixture()
def double_pareto_params():
    alpha = 10
    beta = 1
    return alpha, beta


def get_true_children_params(parent1_params, parent2_params, modified_beta):
    child1_true_params = [((parent1_param + parent2_param) + modified_beta * abs(parent1_param - parent2_param)) / 2
                          for parent1_param, parent2_param in zip(parent1_params, parent2_params)]
    child2_true_params = [((parent1_param + parent2_param) - modified_beta * abs(parent1_param - parent2_param)) / 2
                          for parent1_param, parent2_param in zip(parent1_params, parent2_params)]
    return child1_true_params, child2_true_params


@patch('random.uniform')
@pytest.mark.usefixtures('get_individuals')
def test_double_pareto_1(uniform_mock, get_root_individual, get_individuals, double_pareto_params):
    alpha, beta = double_pareto_params
    uniform_mock.return_value = 0.75
    modified_beta = alpha * beta * (1 - (2 * 0.75) ** (-1 / alpha))
    parent1 = get_individuals[0]
    parent2 = get_individuals[1]
    child1_true_params, child2_true_params = get_true_children_params(parent1.params, parent2.params, modified_beta)
    child1, child2 = DoubleParetoCross.crossover(parent1, parent2, root_individual=get_root_individual,
                                                 dpx_alpha=alpha, dpx_beta=beta)
    assert child1.params == child1_true_params
    assert child2.params == child2_true_params


@patch('random.uniform')
@pytest.mark.usefixtures('get_individuals')
def test_double_pareto_2(uniform_mock, get_root_individual, get_individuals, double_pareto_params):
    alpha, beta = double_pareto_params
    uniform_mock.return_value = 0.25
    modified_beta = alpha * beta * ((1 - (2 * 0.25)) ** (-1 / alpha) - 1)
    parent1, parent2 = get_individuals[0], get_individuals[1]
    child1_true_params, child2_true_params = get_true_children_params(parent1.params, parent2.params, modified_beta)
    child1, child2 = DoubleParetoCross.crossover(parent1, parent2, root_individual=get_root_individual,
                                                 dpx_alpha=alpha, dpx_beta=beta)
    assert child1.params == child1_true_params
    assert child2.params == child2_true_params



#!/usr/bin/env python

# Standard library
from typing import List
import random

# 3rd party packages


# Local source
from domain.selection.strategy import ISelectionStrategy
from domain.individual import Individual


class TournamentSelect(ISelectionStrategy):

    @staticmethod
    def selection(population: List[Individual], **kwargs) -> Individual:
        """Select `tournament_size` individuals from `population` at random and return best individual by fitness.
        Default tournament size = 2.
        """
        tournament_size = kwargs.get('tournament_size', 2)
        selected = random.sample(population, tournament_size)
        return min(selected)



#!/usr/bin/env python

"""Module to transform data into training and test sets."""

# Standard library
from typing import Tuple

# 3rd party packages
import pandas as pd

# Local source


def train_test_split(x_y_df: pd.DataFrame, train_fraction: float) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Split data into training and test sets."""
    train_df = x_y_df.sample(frac=train_fraction)
    test_df = x_y_df.drop(train_df.index)
    return train_df, test_df


def normalize_features(x_df: pd.DataFrame, x_train_stats: pd.DataFrame) -> pd.DataFrame:
    """Normalize the dataset according to training data statistics (obtained from pandas)."""
    return (x_df - x_train_stats['mean']) / x_train_stats['std']


def denormalize_features(x_df: pd.DataFrame, x_train_stats: pd.DataFrame) -> pd.DataFrame:
    """De-normalize the dataset according to the training data statistics (obtained from pandas)."""
    return (x_df * x_train_stats['std']) + x_train_stats['mean']


def get_columns_to_remove(x_df: pd.DataFrame, tolerance: float = 1e-4):
    """Preprocessing step for training ANN. Columns to remove from features dataframe are those with standard
    deviations that are too small and will result in normalization issues.
    """
    return x_df.std()[abs(x_df.std()) <= tolerance].index


def remove_problematic_columns(x_df, columns_to_remove):
    """Return new df with problematic columns removed."""
    return x_df.drop(columns_to_remove, axis=1)



#!/usr/bin/env python

"""Module with helper methods for ffield dict."""

# Standard library
from typing import List, Dict

# 3rd party packages

# Local source


def set_param(key: List[int], ffield: Dict[int, List], param: float):
    """Using a key in the params key list, set the corresponding parameter in the given Child's dictionary.

    :param key: List containing 3 columns: [section # | line/row # | parameter #]
    :param ffield: ffield[section number] = [parameters corresponding to section].
    :param param: Value of the parameter to store in ffield dict at the specified key
    """
    section = key[0]
    row = key[1] - 1  # Indices start at 0 in Python
    ffield[section][row][key[2] - 1] = param


def get_param(key: List[int], ffield: Dict[int, List]) -> float:
    """Using a key in the params key list, find the corresponding value (parameter) in the ffield dictionary.

    :param key: List containing parameter keys that map to the required parameters in the ffield dictionary.
    :param ffield: Dictionary that maps ReaxFF section number to all parameters in that given section.
    :return param: Floating point parameter; value in ffield file at the location of the key.
    """
    section = key[0]
    row = key[1] - 1  # Indices start at 0 in Python
    return ffield[section][row][key[2] - 1]




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.neural_network.extract_data import (individuals_to_features_df,
                                                                      individuals_to_features_and_outputs_df,
                                                                      extract_features_and_outputs_from_combined_df)


@pytest.mark.usefixtures('get_individuals')
def test_individuals_to_features_df(get_individuals):
    features_df = individuals_to_features_df(get_individuals)
    params_nd_array = features_df.to_numpy()
    assert list(params_nd_array[0, :]) == [1.0, 0.1, -0.5, 4.3, 2.4]
    assert list(params_nd_array[1, :]) == [0.5, 0.05, -0.76, 8.6, 2.1]
    assert list(params_nd_array[2, :]) == [1.24, 0.17, -0.07, 6.3, 2.24]
    assert list(params_nd_array[3, :]) == [0.73, 0.25, -0.25, 10.2, 1.7]


@pytest.mark.usefixtures('get_individuals')
def test_individuals_to_features_and_outputs_df(get_individuals):
    features_and_outputs_df = individuals_to_features_and_outputs_df(get_individuals)
    features_and_outputs_nd_array = features_and_outputs_df.to_numpy()
    assert list(features_and_outputs_nd_array[0, :]) == [1.0, 0.1, -0.5, 4.3, 2.4, 43.2, 10.6, -174.2, 1008.4]
    assert list(features_and_outputs_nd_array[1, :]) == [0.5, 0.05, -0.76, 8.6, 2.1, 56.9, 11.2, -164.1, 994.8]
    assert list(features_and_outputs_nd_array[2, :]) == [1.24, 0.17, -0.07, 6.3, 2.24, 102.4, 5.5, -155.5, 1008.12]
    assert list(features_and_outputs_nd_array[3, :]) == [0.73, 0.25, -0.25, 10.2, 1.7, 40.2, 11.2, -104.1, 1018.32]


@pytest.mark.usefixtures('get_individuals')
def test_extract_features_and_outputs_from_combined_df(get_individuals):
    features_and_outputs_df = individuals_to_features_and_outputs_df(get_individuals)
    num_features = len(get_individuals[0].params)
    x_df, y_df = extract_features_and_outputs_from_combined_df(features_and_outputs_df, num_features)
    params_nd_array = x_df.to_numpy()
    reax_energies_nd_array = y_df.to_numpy()
    assert list(params_nd_array[0, :]) == [1.0, 0.1, -0.5, 4.3, 2.4]
    assert list(params_nd_array[1, :]) == [0.5, 0.05, -0.76, 8.6, 2.1]
    assert list(params_nd_array[2, :]) == [1.24, 0.17, -0.07, 6.3, 2.24]
    assert list(params_nd_array[3, :]) == [0.73, 0.25, -0.25, 10.2, 1.7]
    assert list(reax_energies_nd_array[0, :]) == [43.2, 10.6, -174.2, 1008.4]
    assert list(reax_energies_nd_array[1, :]) == [56.9, 11.2, -164.1, 994.8]
    assert list(reax_energies_nd_array[2, :]) == [102.4, 5.5, -155.5, 1008.12]
    assert list(reax_energies_nd_array[3, :]) == [40.2, 11.2, -104.1, 1018.32]



#!/usr/bin/env python

"""Module with interface for selection operators to be performed on Individuals."""

# Standard library
import abc
from typing import List

# 3rd party packages

# Local source
from domain.individual import Individual


class ISelectionStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def selection(population: List[Individual], **kwargs) -> Individual:
        raise NotImplementedError




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.root_individual import RootIndividual
from parametrization_clean.domain.adaptation.srinivas import SrinivasAdapt


@pytest.fixture()
@pytest.mark.usefixtures('dft_energies', 'weights', 'root_ffield', 'param_keys')
def root_individual(dft_energies, weights, root_ffield, param_keys):
    root_individual = RootIndividual(dft_energies, weights, root_ffield, param_keys)
    return root_individual


@pytest.mark.usefixtures('get_individuals')
def test_srinivas_1(get_individuals, root_individual):
    costs = []
    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)
        costs.append(individual.cost)

    average_cost = 500.0
    minimum_cost = min(costs)
    parent_costs = (get_individuals[2].cost, get_individuals[3].cost)
    crossover_rate, mutation_rates = SrinivasAdapt.adaptation(average_cost, minimum_cost, parent_costs)
    assert crossover_rate == 1.0
    assert mutation_rates[0] == 0.5
    assert mutation_rates[1] == 0.5


@pytest.mark.usefixtures('get_individuals')
def test_srinivas_2(get_individuals, root_individual):
    costs = []
    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)
        costs.append(individual.cost)

    average_cost = 1000.0
    minimum_cost = min(costs)
    parent_costs = (get_individuals[2].cost, get_individuals[3].cost)
    crossover_rate, mutation_rates = SrinivasAdapt.adaptation(average_cost, minimum_cost, parent_costs)
    assert crossover_rate == pytest.approx(0.983, rel=1e-3)
    assert mutation_rates[0] == pytest.approx(0.491, rel=1e-3)
    assert mutation_rates[1] == 0.5



#!/usr/bin/env python

"""Factory for crossover algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from domain.crossover.single_point import SinglePointCross
from domain.crossover.two_point import TwoPointCross
from domain.crossover.uniform import UniformCross
from domain.crossover.double_pareto import DoubleParetoCross


def crossover_factory(algorithm_name: str):
    """Factory to select crossover type."""
    crossover_types = {
        'single_point': SinglePointCross,
        'two_point': TwoPointCross,
        'uniform': UniformCross,
        'double_pareto': DoubleParetoCross
    }
    return crossover_types[algorithm_name]




# Standard library
from unittest import mock

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.selection.tournament import TournamentSelect
from parametrization_clean.domain.mutation.gauss import GaussianMutate
from parametrization_clean.domain.crossover.double_pareto import DoubleParetoCross
from parametrization_clean.domain.adaptation.xiao import XiaoAdapt
from parametrization_clean.domain.cost.reax_error import ReaxError

# Fixtures
from tests.use_case.test_population_propagator import root_individual


@pytest.fixture()
@mock.patch('parametrization_clean.use_case.port.settings_repository.IAllSettings')
@pytest.mark.usefixtures('param_bounds')
def all_settings(all_settings_mock, param_bounds):
    all_settings_mock.strategy_settings.selection_strategy = TournamentSelect
    all_settings_mock.strategy_settings.mutation_strategy = GaussianMutate
    all_settings_mock.strategy_settings.crossover_strategy = DoubleParetoCross
    all_settings_mock.strategy_settings.adaptation_strategy = XiaoAdapt
    all_settings_mock.strategy_settings.error_strategy = ReaxError

    all_settings_mock.ga_settings.population_size = 4
    all_settings_mock.ga_settings.mutation_rate = 0.20
    all_settings_mock.ga_settings.crossover_rate = 0.80
    all_settings_mock.ga_settings.use_elitism = True
    all_settings_mock.ga_settings.use_adaptation = False

    all_settings_mock.mutation_settings.gauss_std = [0.10]
    all_settings_mock.mutation_settings.gauss_frac = [1.0]
    all_settings_mock.mutation_settings.nakata_rand_lower = -1.0
    all_settings_mock.mutation_settings.nakata_rand_higher = 1.0
    all_settings_mock.mutation_settings.nakata_scale = 0.10
    all_settings_mock.mutation_settings.polynomial_eta = 60
    all_settings_mock.mutation_settings.param_bounds = param_bounds

    all_settings_mock.crossover_settings.dpx_alpha = 10
    all_settings_mock.crossover_settings.dpx_beta = 1

    all_settings_mock.selection_settings.tournament_size = 2

    all_settings_mock.adaptation_settings.srinivas_k1 = 1.0
    all_settings_mock.adaptation_settings.srinivas_k2 = 0.5
    all_settings_mock.adaptation_settings.srinivas_k3 = 1.0
    all_settings_mock.adaptation_settings.srinivas_k4 = 0.5
    all_settings_mock.adaptation_settings.srinivas_default_mutation_rate = 0.005
    all_settings_mock.adaptation_settings.xiao_min_crossover_rate = 0.4
    all_settings_mock.adaptation_settings.xiao_min_mutation_rate = 0.1
    all_settings_mock.adaptation_settings.xiao_scale = 0.4

    all_settings_mock.neural_net_settings.verbosity = 2
    all_settings_mock.neural_net_settings.train_fraction = 0.80
    all_settings_mock.neural_net_settings.num_epochs = 1
    all_settings_mock.neural_net_settings.num_populations_to_train_on = 10
    all_settings_mock.neural_net_settings.num_nested_ga_iterations = 2
    all_settings_mock.neural_net_settings.minimum_validation_r_squared = 0.95

    return all_settings_mock


@pytest.mark.usefixtures('get_individuals')
@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_genetic_neural_net_propagator_init(repository_mock, all_settings, get_individuals, root_individual):
    # Conditional import machinery
    nested_ga_with_ann = pytest.importorskip('parametrization_clean.use_case.nested_ga_with_ann')

    repository_mock.get_previous_n_populations = mock.MagicMock(return_value=get_individuals*10)
    repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    propagator = nested_ga_with_ann.GeneticNeuralNetPropagator(all_settings, repository_mock)

    assert propagator.neural_net_settings.verbosity == 2
    assert propagator.neural_net_settings.train_fraction == 0.80
    assert propagator.neural_net_settings.num_epochs == 1
    assert propagator.neural_net_settings.num_populations_to_train_on == 10
    assert propagator.neural_net_settings.num_nested_ga_iterations == 2
    assert propagator.neural_net_settings.minimum_validation_r_squared == 0.95

    assert propagator.root_individual == root_individual

    assert propagator.error_strategy == ReaxError
    assert propagator.population_size == 4


@pytest.mark.usefixtures('get_individuals')
@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_genetic_neural_net_propagator_train_neural_net(repository_mock, all_settings,
                                                        get_individuals, root_individual):
    # Conditional import machinery
    nested_ga_with_ann = pytest.importorskip('parametrization_clean.use_case.nested_ga_with_ann')

    repository_mock.get_previous_n_populations = mock.MagicMock(return_value=get_individuals)
    repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    propagator = nested_ga_with_ann.GeneticNeuralNetPropagator(all_settings, repository_mock)

    # Just see if the model and history are being created to see if neural net is actually being trained
    model, history = propagator.train_neural_net()
    assert model
    assert history


@pytest.mark.usefixtures('get_individuals')
@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_genetic_neural_net_propagator_update_costs(repository_mock, all_settings,
                                                    get_individuals, root_individual):
    # Conditional import machinery
    nested_ga_with_ann = pytest.importorskip('parametrization_clean.use_case.nested_ga_with_ann')

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)

    repository_mock.get_previous_n_populations = mock.MagicMock(return_value=get_individuals)
    repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    propagator = nested_ga_with_ann.GeneticNeuralNetPropagator(all_settings, repository_mock)

    model, _ = propagator.train_neural_net()
    next_generation, _ = propagator.propagate_first(get_individuals, model)
    y_predicted = propagator.neural_net.predict_outputs(model, next_generation)
    propagator.update_costs(next_generation, y_predicted)

    old_costs = []
    new_costs = []
    for old_individual, new_individual in zip(get_individuals, next_generation):
        old_costs.append(old_individual.cost)
        new_costs.append(new_individual.cost)

    assert len(new_costs) == 4
    assert len(y_predicted) == 4
    assert len(y_predicted[0]) == 4
    for new_cost in new_costs:
        assert new_cost


@pytest.mark.usefixtures('get_individuals')
@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_genetic_neural_net_propagator_propagate_first(repository_mock, all_settings,
                                                       get_individuals, root_individual):
    # Conditional import machinery
    nested_ga_with_ann = pytest.importorskip('parametrization_clean.use_case.nested_ga_with_ann')

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)

    repository_mock.get_previous_n_populations = mock.MagicMock(return_value=get_individuals)
    repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    propagator = nested_ga_with_ann.GeneticNeuralNetPropagator(all_settings, repository_mock)

    model, _ = propagator.train_neural_net()
    next_generation, best_master_parents = propagator.propagate_first(get_individuals, model)
    best_costs = [best_master_parents[0].cost, best_master_parents[1].cost]
    assert next_generation != get_individuals
    assert pytest.approx(239.557, rel=1e-3) in best_costs
    assert pytest.approx(871.989, rel=1e-3) in best_costs


@pytest.mark.usefixtures('get_individuals')
@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_genetic_neural_net_propagator_propagate_remaining(repository_mock, all_settings,
                                                           get_individuals, root_individual):
    # Conditional import machinery
    nested_ga_with_ann = pytest.importorskip('parametrization_clean.use_case.nested_ga_with_ann')

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)

    repository_mock.get_previous_n_populations = mock.MagicMock(return_value=get_individuals)
    repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    propagator = nested_ga_with_ann.GeneticNeuralNetPropagator(all_settings, repository_mock)

    model, _ = propagator.train_neural_net()
    next_generation, _ = propagator.propagate_first(get_individuals, model)
    final_generation = propagator.propagate_remaining(next_generation, model)
    assert len(final_generation) == 4


@pytest.mark.usefixtures('get_individuals')
@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_genetic_neural_propagator_final_ann_accuracy_is_poor(repository_mock, all_settings,
                                                              get_individuals, root_individual):
    # Conditional import machinery
    nested_ga_with_ann = pytest.importorskip('parametrization_clean.use_case.nested_ga_with_ann')

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)

    repository_mock.get_previous_n_populations = mock.MagicMock(return_value=get_individuals)
    repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    propagator = nested_ga_with_ann.GeneticNeuralNetPropagator(all_settings, repository_mock)

    _, history = propagator.train_neural_net()
    assert propagator.final_ann_accuracy_is_poor(history)
    history.history['val_r_square'][-1] = 0.96
    assert not propagator.final_ann_accuracy_is_poor(history)


@pytest.mark.usefixtures('get_individuals')
@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_genetic_neural_net_propagator_execute(repository_mock, all_settings,
                                               get_individuals, root_individual):
    # Conditional import machinery
    nested_ga_with_ann = pytest.importorskip('parametrization_clean.use_case.nested_ga_with_ann')

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)

    repository_mock.get_previous_n_populations = mock.MagicMock(return_value=get_individuals)
    repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    propagator = nested_ga_with_ann.GeneticNeuralNetPropagator(all_settings, repository_mock)

    final_generation, model, history = propagator.execute(get_individuals)
    assert len(final_generation) == 4
    assert model
    assert history



#!/usr/bin/env python

"""Module to extract Individual's data to a convenient form for neural network training."""

# Standard library
from typing import List, Tuple

# 3rd party packages
import pandas as pd

# Local source
from domain.individual import Individual


def individuals_to_features_df(population: List[Individual]) -> pd.DataFrame:
    return pd.DataFrame(data=[individual.params for individual in population])


def individuals_to_features_and_outputs_df(population: List[Individual]) -> pd.DataFrame:
    params_and_reax_predictions = [individual.params + individual.reax_energies for individual in population]
    return pd.DataFrame(data=params_and_reax_predictions)


def extract_features_and_outputs_from_combined_df(x_y_df: pd.DataFrame, num_features: int) \
        -> Tuple[pd.DataFrame, pd.DataFrame]:
    x_df = x_y_df.iloc[:, 0:num_features]
    y_df = x_y_df.iloc[:, num_features:]
    return x_df, y_df



#!/usr/bin/env python

"""
Module that contains interface for repository used to get data on Individual(s) in the genetic pool.

__author__ = "Chad Daksha"
"""

# Standard library
import abc
from typing import List

# 3rd party packages


# Local source
from domain.root_individual import RootIndividual
from domain.individual import Individual


class IPopulationRepository(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_root_individual(self) -> RootIndividual:
        """Get RootIndividual data from reference training set."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_population(self, generation_number: int) -> List[Individual]:
        """Get population of Individuals based on the (unique) generation number."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_previous_n_populations(self, num_populations: int) -> List[Individual]:
        """Read previous N populations relative to the current generation.
        Precaution - ensure that lowest possible generation number is zero.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def write_individual(self, individual: Individual, **kwargs):
        """Write an individual in the population to the repository."""
        raise NotImplementedError

    @abc.abstractmethod
    def write_population(self, population: List[Individual], generation_number):
        """Write a population of individuals to the repository."""
        raise NotImplementedError



#!/usr/bin/env python

"""Build script for setuptools. Tells setuptools about your package (such as the name and version)
as well as which code files to include.
"""

from setuptools import setup, find_packages

with open("README.md", 'r') as readme_file:
    readme = readme_file.read()

requirements = []

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Chad Daksha",
    author_email='daksha@udel.edu',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
    description="ReaxFF parameter optimization scheme using generational genetic algorithm and neural networks.",
    entry_points={
        'console_scripts': [
            'cli=parametrization_clean.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='parametrization_clean',
    name='parametrization_clean-cdaksha',
    packages=find_packages(include=['parametrization_clean', 'parametrization_clean.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/cdaksha/parametrization_clean',
    version='1.0.0.3',
    zip_safe=False,
)




# Standard library

# 3rd party packages

# Local source
from domain.selection.factory import selection_factory
from domain.selection.tournament import TournamentSelect


def test_get_tournament():
    assert selection_factory('tournament') == TournamentSelect



#!/usr/bin/env python

# Standard library
from typing import Tuple

# 3rd party packages
import numpy as np

# Local source
from domain.crossover.strategy import ICrossoverStrategy
from domain.individual import Individual
from domain.root_individual import RootIndividual


class UniformCross(ICrossoverStrategy):

    @staticmethod
    def crossover(parent1: Individual, parent2: Individual, root_individual: RootIndividual,
                  **kwargs) -> Tuple[Individual, Individual]:
        """Execute uniform crossover. Take the ith row of parent1 and randomly swap bits with the ith row of p2."""
        sieve = np.random.randint(2, size=len(parent1.params))  # Array of 0's and 1's
        not_sieve = sieve ^ 1  # Complement of sieve

        child1 = Individual(list(parent1.params * sieve + parent2.params * not_sieve), root_individual=root_individual)
        child2 = Individual(list(parent1.params * not_sieve + parent2.params * sieve), root_individual=root_individual)

        return child1, child2



#!/usr/bin/env python

"""Main module to initialize/propagate genetic population and output files containing Individuals to execute
using ReaxFF.
"""

# Standard library

# 3rd party packages

# Local source
from use_case.population_initializer import PopulationInitializer
from use_case.population_propagator import PopulationPropagator
from use_case.population_writer import PopulationWriter
from infrastructure.config.local import UserSettings
from infrastructure.repository.from_files import PopulationFileRepository
from infrastructure.presenter.file_writer import DataWriter


def run_application(generation_number, training_path, population_path):
    user_settings = UserSettings()
    population_repository = PopulationFileRepository(training_path, population_path,
                                                     user_settings, generation_number)
    population_writer = PopulationWriter(population_repository)

    if generation_number == 1:
        # First generation of the genetic algorithm --> initialize first population
        population_initializer = PopulationInitializer(population_repository, user_settings)
        next_population = population_initializer.execute()
    else:
        # Generation number > 1. Propagate population by either using standalone GA or by using GA + nested ANN.
        previous_generation_number = generation_number - 1
        previous_population, successfully_retrieved_case_numbers = \
            population_repository.get_population(previous_generation_number)

        enough_generations_elapsed = generation_number >= user_settings.neural_net_settings.num_populations_to_train_on
        use_neural_network = user_settings.ga_settings.use_neural_network
        if use_neural_network and enough_generations_elapsed:
            from parametrization_clean.use_case.nested_ga_with_ann import GeneticNeuralNetPropagator
            population_propagator = GeneticNeuralNetPropagator(user_settings, population_repository)
            next_population, _, history = population_propagator.execute(previous_population)
        else:
            population_propagator = PopulationPropagator(user_settings, population_repository)
            next_population = population_propagator.execute(previous_population)
            history = None

        DataWriter.write_outputs(previous_population, successfully_retrieved_case_numbers, population_repository,
                                 user_settings, generation_number, history)

    response = population_writer.write_population(next_population, generation_number)

    return response




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.mutation.polynomial import PolynomialMutate
from tests.domain.crossover.test_double_pareto import get_root_individual


@pytest.mark.usefixtures('get_individuals')
def test_polynomial(get_root_individual, get_individuals):
    parent = get_individuals[0]
    child = PolynomialMutate.mutation(parent, get_root_individual, polynomial_eta=60)
    assert parent.params != child.params
    assert parent.params == get_individuals[0].params



#!/usr/bin/env python

"""Module with immutable structure to store reference training set information.
Potentially includes DFT energies, weights, parameter boundaries (low, high)."""

# Standard library
from typing import List, Dict


# 3rd party packages

# Local source
from domain.utils.helpers import get_param


class Borg(object):
    """Avoid Singleton anti-pattern, but ensure shared state (if many instances are created)."""
    _shared_state = {}

    def __init__(self):
        self.__dict__ = self._shared_state


class RootIndividual(Borg):

    def __init__(self, dft_energies: List[float], weights: List[float],
                 root_ffield: Dict[int, List], param_keys: List[List[int]]):
        """Root individual to store data from reference training set that needs to be stored only once.
        For example, the weights of the error, DFT energies, parameter bounds only need to be stored once.
        Input structures are converted to tuples to prevent mutation.

        :param dft_energies: Tuple containing DFT energies, used to compute error.
        :param weights: Tuple containing weights, used to compute error.
        :param root_ffield: Dictionary mapping section number to data contained in that section.
        :param param_keys: List of keys pointing to values in the ffield object.
        """
        super().__init__()
        self.dft_energies = tuple(dft_energies)
        self.weights = tuple(weights)
        self.root_ffield = root_ffield
        self.param_keys = param_keys

        # Root parameters from the reference training set
        self.root_params = self.extract_params()

    def extract_params(self) -> List[float]:
        return [get_param(key, self.root_ffield) for key in self.param_keys]



#!/usr/bin/env python

# Standard library
from typing import Tuple, List

# 3rd party packages


# Local source
from domain.adaptation.strategy import IAdaptationStrategy


class SrinivasAdapt(IAdaptationStrategy):

    @staticmethod
    def adaptation(average_cost: float, minimum_cost: float, parent_costs: Tuple[float, float], **kwargs) \
            -> Tuple[float, List[float]]:
        """Execute Srinivas adaptive algorithm.

        :param average_cost: Measure of central tendency of cost of population. Can also be the median.
        :param minimum_cost: Lowest cost of population of Individuals.
        :param parent_costs: Tuple containing costs associated with two parents.
        :return: Tuple with (probability of crossover, [probabilities of mutation]).
        """
        k1 = kwargs.get('srinivas_k1', 1.0)
        k2 = kwargs.get('srinivas_k2', 0.5)
        k3 = kwargs.get('srinivas_k3', 1.0)
        k4 = kwargs.get('srinivas_k4', 0.5)
        default_mutation_rate = kwargs.get('srinivas_default_mutation_rate', 0.005)

        spread = average_cost - minimum_cost

        better_parent_cost = min(parent_costs)
        if better_parent_cost <= average_cost:
            cross_rate = k1 * (better_parent_cost - minimum_cost) / spread
        else:
            cross_rate = k3

        mutation_rates = []
        for cost in parent_costs:
            if cost <= average_cost:
                new_mutation_rate = max(default_mutation_rate, k2 * (cost - minimum_cost) / spread)
                mutation_rates.append(new_mutation_rate)
            else:
                mutation_rates.append(k4)

        return cross_rate, mutation_rates



#!/usr/bin/env python

"""Tests for `parametrization_clean` package."""

import os
import shutil

import pytest
import click
from click.testing import CliRunner

from parametrization_clean.cli import main


@pytest.mark.usefixtures("training_set_dir_path", "cli_output_path", "reax_output_dir_path")
def test_command_line_interface(training_set_dir_path, cli_output_path, reax_output_dir_path):
    """Test the CLI."""
    training_path = str(training_set_dir_path)
    population_path = str(cli_output_path)
    generation_number = 1

    runner = CliRunner()
    result = runner.invoke(main, '--generation_number {} --training_path="{}" --population_path="{}"'
                           .format(generation_number, training_path, population_path))

    assert result.exit_code == 0
    assert "Generation Number: 1\n" in result.output
    assert "Retrieving reference data from: {}".format(training_path) in result.output
    assert "Outputting generational genetic algorithm data to: {}".format(population_path) in result.output
    assert "{}: {}".format("SUCCESS", "Generation successfully written at {}".format(population_path)) in result.output
    assert not result.exception

    help_result = runner.invoke(main, ['--help'])
    assert help_result.exit_code == 0
    assert "Usage: cli [OPTIONS]" in help_result.output
    assert "Command-line interface for genetic algorithm + neural network generational\n  propagation" in help_result.output
    assert "-g, --generation_number" in help_result.output
    assert "-t, --training_path" in help_result.output
    assert "-p, --population_path" in help_result.output
    assert not help_result.exception

    # Teardown
    shutil.rmtree(os.path.join(population_path, "generation-1"))

    population_path = str(reax_output_dir_path)
    result = runner.invoke(main, '--generation_number {} --training_path="{}" --population_path="{}"'
                           .format(3, training_path, population_path))
    assert result.exit_code == 0
    assert "Generation Number: 3\n" in result.output
    assert "Retrieving reference data from: {}".format(training_path) in result.output
    assert "Outputting generational genetic algorithm data to: {}".format(population_path) in result.output
    assert "{}: {}".format("SUCCESS", "Generation successfully written at {}".format(population_path)) in result.output
    assert not result.exception

    # Teardown
    shutil.rmtree(os.path.join(population_path, "generation-3"))

    result = runner.invoke(main, '')
    assert result.exception



#!/usr/bin/env python

"""
Module that contains interface for crossover methods to be used for an Individual.

__author__ = "Chad Daksha"
"""

# Standard library
import abc
from typing import Tuple

# 3rd party packages


# Local source
from domain.individual import Individual
from domain.root_individual import RootIndividual


class ICrossoverStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def crossover(parent1: Individual, parent2: Individual, root_individual: RootIndividual,
                  **kwargs) -> Tuple[Individual, Individual]:
        raise NotImplementedError



#!/usr/bin/env python

# Standard library
from typing import Tuple
import random

# 3rd party packages

# Local source
from domain.crossover.strategy import ICrossoverStrategy
from domain.individual import Individual
from domain.root_individual import RootIndividual


class DoubleParetoCross(ICrossoverStrategy):

    @staticmethod
    def crossover(parent1: Individual, parent2: Individual, root_individual: RootIndividual,
                  **kwargs) -> Tuple[Individual, Individual]:
        """Double Pareto crossover from Thakur's 2014 - "A new GA for global optimization of multimodal continuous
        functions."
        NOTE: Thakur is unclear about modified beta if/then command.
        I have ASSUMED the following:
        if u >= 1/2:
            >> use first expression
        else: (u < 1/2):
            >> use second expression
        This SHOULD be correct, as x = 0 corresponds to f(x), which is the Pareto density function, equal to 0.5!
        """
        alpha = kwargs.get('dpx_alpha', 10)
        beta = kwargs.get('dpx_beta', 1)

        child1_params = []
        child2_params = []
        for parent1_param, parent2_param in zip(parent1.params, parent2.params):
            u = random.uniform(0, 1)
            if u >= 1 / 2:
                modified_beta = alpha * beta * (1 - (2 * u) ** (-1 / alpha))
            else:
                modified_beta = alpha * beta * ((1 - (2 * u)) ** (-1 / alpha) - 1)

            child1_param = ((parent1_param + parent2_param) + modified_beta * abs(parent1_param - parent2_param)) / 2
            child2_param = ((parent1_param + parent2_param) - modified_beta * abs(parent1_param - parent2_param)) / 2

            child1_params.append(child1_param)
            child2_params.append(child2_param)

        child1 = Individual(child1_params, root_individual=root_individual)
        child2 = Individual(child2_params, root_individual=root_individual)
        return child1, child2



#!/usr/bin/env python

"""Present data from results of genetic algorithm/ANN."""

# Standard library
import csv
import os

# 3rd party packages

# Local source


class DataWriter:

    @staticmethod
    def write_rows_to_csv(file_path, data, delimiter=','):
        with open(file_path, 'w', newline='') as out_file:
            writer = csv.writer(out_file, delimiter=delimiter)
            writer.writerows(data)

    @staticmethod
    def create_or_append(formatted_data_string, file_path):
        with open(file_path, 'a+') as out_file:
            out_file.write(formatted_data_string)

    @staticmethod
    def write_summary(population, expected_population_size, successfully_retrieved_case_numbers, file_path):
        """Generate report for a given generation at location `file_path`."""
        best_child = sorted(population)[0]

        num_retrieved = len(population)
        num_failed = expected_population_size - num_retrieved

        best_child_idx = population.index(best_child)
        with open(file_path, 'w') as out_file:
            out_file.write("---------- GENERATION SUMMARY ----------\n")
            out_file.write("Number of successful cases: {:.3f} ({:.2f}%)\n"
                           .format(num_retrieved, 100 * num_retrieved / expected_population_size))
            out_file.write("Number of failed cases: {:.3f} ({:.2f}%)\n"
                           .format(num_failed, 100 * num_failed / expected_population_size))
            out_file.write("Best Cost/Fitness Real ReaxFF Error (from Master GA): {:.3f}\n"
                           .format(best_child.cost))
            out_file.write("Corresponding best Master GA case: case-{}\n".
                           format(successfully_retrieved_case_numbers[best_child_idx]))

    @staticmethod
    def write_outputs(previous_population, successfully_retrieved_case_numbers, population_repository, user_settings,
                      generation_number, history=None):
        """Write summary of previous generation, append to generation # vs. best total error,
        and append to ANN summary.
        """
        previous_generation_number = generation_number - 1
        population_path = population_repository.population_path
        expected_population_size = user_settings.ga_settings.population_size
        if history:
            history_dict = history.history

        summary_file_path = os.path.join(population_path, 'generation-' + str(previous_generation_number),
                                         '00-gen-summary.txt')
        DataWriter.write_summary(previous_population, expected_population_size,
                                 successfully_retrieved_case_numbers, summary_file_path)

        generation_vs_error_file_path = os.path.join(population_path, '00-generation-vs-error.txt')
        neural_network_file_path = os.path.join(population_path, "00-ann-summary")
        if previous_generation_number == 1:
            # First generation -> initialize files
            DataWriter.create_or_append("Generation Number, Best Total Error\n", generation_vs_error_file_path)
            if history:
                DataWriter.create_or_append("Generation #\t{}\n"
                                            .format(list(history_dict.keys())), neural_network_file_path)

        best_previous_error = sorted(previous_population)[0].cost
        DataWriter.create_or_append("{},{:.3f}\n".format(previous_generation_number, best_previous_error),
                                    generation_vs_error_file_path)

        if history:
            final_neural_network_results = [val[-1] for val in history_dict.values()]
            num_columns = len(final_neural_network_results)
            DataWriter.create_or_append(('{}\t' + '{:.3f}\t' * num_columns + '\n').format(
                previous_generation_number, *final_neural_network_results), neural_network_file_path)



#!/usr/bin/env python

"""
Module that contains interface for repository used to get configuration settings.

__author__ = "Chad Daksha"
"""

# Standard library
import abc
from typing import List

# 3rd party packages

# Local source
from domain.selection.strategy import ISelectionStrategy
from domain.crossover.strategy import ICrossoverStrategy
from domain.mutation.strategy import IMutationStrategy
from domain.adaptation.strategy import IAdaptationStrategy
from domain.cost.strategy import IErrorStrategy


class IStrategySettings(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def selection_strategy(self) -> ISelectionStrategy:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def mutation_strategy(self) -> IMutationStrategy:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def crossover_strategy(self) -> ICrossoverStrategy:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def adaptation_strategy(self) -> IAdaptationStrategy:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def error_strategy(self) -> IErrorStrategy:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def initialization_strategy(self) -> IMutationStrategy:
        raise NotImplementedError


class IGeneticAlgorithmSettings(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def population_size(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def mutation_rate(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def crossover_rate(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def use_elitism(self) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def use_adaptation(self) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def use_neural_network(self) -> bool:
        raise NotImplementedError


class IMutationSettings(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def gauss_std(self) -> List[float]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def gauss_frac(self) -> List[float]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def nakata_rand_lower(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def nakata_rand_higher(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def nakata_scale(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def polynomial_eta(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def param_bounds(self):
        raise NotImplementedError


class ICrossoverSettings(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def dpx_alpha(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dpx_beta(self) -> float:
        raise NotImplementedError


class ISelectionSettings(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def tournament_size(self) -> int:
        raise NotImplementedError


class IAdaptationSettings(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def srinivas_k1(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def srinivas_k2(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def srinivas_k3(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def srinivas_k4(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def srinivas_default_mutation_rate(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def xiao_min_crossover_rate(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def xiao_min_mutation_rate(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def xiao_scale(self) -> float:
        raise NotImplementedError


class INeuralNetSettings(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def verbosity(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def train_fraction(self) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def num_epochs(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def num_populations_to_train_on(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def num_nested_ga_iterations(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def minimum_validation_r_squared(self) -> float:
        raise NotImplementedError


class IAllSettings(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def strategy_settings(self) -> IStrategySettings:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def ga_settings(self) -> IGeneticAlgorithmSettings:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def mutation_settings(self) -> IMutationSettings:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def crossover_settings(self) -> ICrossoverSettings:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def selection_settings(self) -> ISelectionSettings:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def adaptation_settings(self) -> IAdaptationSettings:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def neural_net_settings(self) -> INeuralNetSettings:
        raise NotImplementedError
